﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2013 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Analysis.SolutionCaching.RealVectorEncoding;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.TestFunctions;
using Mehroz;

namespace HeuristicLab.Analysis.AlgorithmBehavior.Analyzers {
  [Item("RealVectorConvexHullModifier", "Calculates the convex hull from a solution cache. ")]
  [StorableClass]
  public class RealVectorConvexHullModifier : ParameterizedNamedItem, IRunCollectionModifier {
    private const string SolutionCacheResultName = "SolutionCache";
    private const string CentroidMovementDistancesName = "Centroid Movement Distances";
    private const string CentroidMotionName = "Centroid Motion";
    private const string BestSolutionName = "Best Solution";
    private const string IncrementalHullVolumeRatioName = "Incremental Hull Volume Ratio";

    public IValueParameter<DoubleArray> BestSolutionParameter {
      get { return (ValueParameter<DoubleArray>)Parameters[BestSolutionName]; }
    }

    [StorableConstructor]
    protected RealVectorConvexHullModifier(bool deserializing) : base(deserializing) { }
    protected RealVectorConvexHullModifier(RealVectorConvexHullModifier original, Cloner cloner) : base(original, cloner) { }
    public RealVectorConvexHullModifier() {
      Parameters.Add(new ValueParameter<DoubleArray>(BestSolutionName, new DoubleArray()));
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new RealVectorConvexHullModifier(this, cloner);
    }

    public void Modify(List<IRun> runs) {
      foreach (var run in runs) {
        int i = 0;
        if (!run.Results.ContainsKey(SolutionCacheResultName)) continue;
        var solutionCache = run.Results.Single(x => x.Key == "SolutionCache").Value as RealVectorSolutionCache;
        if (solutionCache == null) continue;

        List<double[]> centroidList = new List<double[]>();
        var bestSolution = run.Results["Best Solution"] as SingleObjectiveTestFunctionSolution;
        var bestKnownSolution = BestSolutionParameter.Value;

        DataTable volDataTable = new DataTable("Convex hull volume over generations");
        DataTable nrOfPointsTable = new DataTable("Nr. of points of convex hull");
        DataTable centroidDistancesTable = new DataTable(CentroidMovementDistancesName);
        DataTable centroidMotionTable = new DataTable(CentroidMotionName);
        DataTable incrementalHullVolumeRatioTable = new DataTable(IncrementalHullVolumeRatioName);
        DoubleValue overallVolume = new DoubleValue();
        DoubleValue overallVolumeRatio = new DoubleValue();

        DataRow dtVolumeRow = new DataRow("Volume");
        DataRow dtNrPointsRow = new DataRow("Nr. of points");
        DataRow dtCentroidDistances = new DataRow("Distances");
        DataRow dtCentroidMotion = new DataRow("Motion");
        DataRow dtCentroidDistanceFromOptimum = new DataRow("Distance from optimum");
        DataRow dtIncrementalHullVolume = new DataRow("Incremental Hull Volume");
        DataRow dtIncrementalHullVolumeRatio = new DataRow("Incremental Hull Volume Ratio");
        centroidMotionTable.Rows.Add(dtCentroidMotion);
        centroidDistancesTable.Rows.Add(dtCentroidDistances);
        centroidDistancesTable.Rows.Add(dtCentroidDistanceFromOptimum);
        volDataTable.Rows.Add(dtVolumeRow);
        volDataTable.Rows.Add(dtIncrementalHullVolume);
        nrOfPointsTable.Rows.Add(dtNrPointsRow);
        incrementalHullVolumeRatioTable.Rows.Add(dtIncrementalHullVolumeRatio);

        var boundsMatrix = run.Parameters["Bounds"] as DoubleMatrix;
        var bounds = new double[] { boundsMatrix[0, 0], boundsMatrix[0, 1] };
        Fraction hyperCubeVolumeBig = ConvexHullMeasures.CalculateHypercubeVolumeBig(bounds, BestSolutionParameter.Value.Length);
        double hyperCubeVolume = double.NaN;

        if (hyperCubeVolumeBig < 10000000) {
          hyperCubeVolume = ConvexHullMeasures.CalculateHypercubeVolume(bounds, BestSolutionParameter.Value.Length);
        }

        List<double[]> completeHull = null;
        var sols = solutionCache.GetSolutionsFromGeneration(i);
        while (sols.Count != 0) {
          var input = sols.Select(x => ConvertRealVectorToVertexArray(x)).ToArray();
          var convexHull = LPHull.Calculate(input);
          if (convexHull.First().Length < convexHull.Count) {
            var volume = ConvexHullMeasures.CalculateVolume(convexHull);
            dtVolumeRow.Values.Add(volume);
          } else {
            dtVolumeRow.Values.Add(double.NaN);
          }
          dtNrPointsRow.Values.Add(convexHull.Count);
          centroidList.Add(ConvexHullMeasures.CalculateCentroids(convexHull));

          //incrementally build complete convex hull
          if (completeHull == null) {
            completeHull = convexHull;
          } else {
            var newPHull = completeHull.Union(convexHull).ToArray();
            completeHull = LPHull.Calculate(newPHull);
          }

          if (completeHull != null && completeHull.Any() && completeHull.First().Length < completeHull.Count) {
            if (double.IsNaN(hyperCubeVolume)) {
              Fraction completeHullVolume = ConvexHullMeasures.CalculateVolumeBig(completeHull);
              Fraction ratio = completeHullVolume / hyperCubeVolumeBig;
              dtIncrementalHullVolume.Values.Add(completeHullVolume.ToDouble());
              dtIncrementalHullVolumeRatio.Values.Add(ratio.ToDouble());
            } else {
              double completeHullVolume = ConvexHullMeasures.CalculateVolume(completeHull);
              double ratio = completeHullVolume / hyperCubeVolume;
              dtIncrementalHullVolume.Values.Add(completeHullVolume);
              dtIncrementalHullVolumeRatio.Values.Add(ratio);
            }
          }

          sols = solutionCache.GetSolutionsFromGeneration(++i);
        }

        if (completeHull != null && completeHull.Any() && completeHull.First().Length < completeHull.Count) {
          if (!double.IsNaN(hyperCubeVolume)) {
            overallVolume.Value = ConvexHullMeasures.CalculateVolume(completeHull);
            overallVolumeRatio.Value = overallVolume.Value / hyperCubeVolume;
          } else {
            Fraction v = ConvexHullMeasures.CalculateVolumeBig(completeHull);
            overallVolume.Value = v.ToDouble();
            overallVolumeRatio.Value = (double)(v / hyperCubeVolumeBig);
          }
        } else {
          overallVolume.Value = double.NaN;
          overallVolumeRatio.Value = double.NaN;
        }

        CalculateMovementDistances(dtCentroidDistances, centroidList, run);
        CalculateMotionMeasures(dtCentroidMotion, centroidList, run);
        CalculateCentroidDistanceFromOptimum(dtCentroidDistanceFromOptimum, centroidList, bestKnownSolution.ToArray());

        run.Results["Overall volume"] = overallVolume;
        run.Results["Overall volume ratio"] = overallVolumeRatio;
        run.Results["Overall diameter"] = new DoubleValue(ConvexHullMeasures.CalculateMaxDiameter(completeHull));
        run.Results["Convex hull volume"] = volDataTable;
        run.Results["Nr. of points on convex hull"] = nrOfPointsTable;
        run.Results[CentroidMovementDistancesName] = centroidDistancesTable;
        run.Results[CentroidMotionName] = centroidMotionTable;
        run.Results[IncrementalHullVolumeRatioName] = incrementalHullVolumeRatioTable;
      }
    }

    private void CalculateCentroidDistanceFromOptimum(DataRow row, List<double[]> centroidList, double[] bestKnownSolution) {
      foreach (var centroid in centroidList) {
        double distance = centroid.EuclideanDistance(bestKnownSolution);
        row.Values.Add(distance);
      }
    }

    private void CalculateMovementDistances(DataRow row, List<double[]> centroidList, IRun run) {
      var distances = ConvexHullMeasures.CalculateMovementDistances(centroidList);
      foreach (var distance in distances) {
        row.Values.Add(distance);
      }
      run.Results["Centroid Movement Distance Avg."] = new DoubleValue(distances.Average());
      run.Results["Centroid Movement Distance Std.Dev."] = new DoubleValue(distances.StandardDeviation());
      run.Results["Centroid Movement Distance Sum"] = new DoubleValue(distances.Sum());
      run.Results["Centroid Movement Overall Distance"] = new DoubleValue(ConvexHullMeasures.CalculateOverallMovementDistances(centroidList));
    }

    private void CalculateMotionMeasures(DataRow row, List<double[]> centroidList, IRun run) {
      var motions = ConvexHullMeasures.CalculateCentroidsMotion(centroidList).Select(y => Math.Abs(y)).ToList();
      foreach (var motion in motions) {
        row.Values.Add(motion);
      }
      run.Results["Centroid Motion Avg."] = new DoubleValue(motions.Average());
      run.Results["Centroid Motion Std.Dev."] = new DoubleValue(motions.StandardDeviation());
      run.Results["Centroid Motion Sum"] = new DoubleValue(motions.Sum());
    }

    private double[] ConvertRealVectorToVertexArray(RealVector p) {
      double[] vertex = p.Select(x => (double)x).ToArray();
      return vertex;
    }
  }
}